
#include "rtsp-media-factory-custom.h"

enum
{
  PROP_0,
};

static void (*parent_finalize)(GObject* obj) = NULL;

GST_DEBUG_CATEGORY_STATIC (rtsp_media_debug);
#define GST_CAT_DEFAULT rtsp_media_debug

static void gst_rtsp_media_factory_custom_get_property (GObject * object, guint propid,
    GValue * value, GParamSpec * pspec);
static void gst_rtsp_media_factory_custom_set_property (GObject * object, guint propid,
    const GValue * value, GParamSpec * pspec);
static void gst_rtsp_media_factory_custom_finalize (GObject * obj);

static gchar *custom_gen_key (GstRTSPMediaFactory * factory,
    const GstRTSPUrl * url);
static GstElement *custom_get_element (GstRTSPMediaFactory * factory,
    const GstRTSPUrl * url);
static GstRTSPMedia *custom_construct (GstRTSPMediaFactory * factory,
    const GstRTSPUrl * url);
static void custom_configure (GstRTSPMediaFactory * factory,
    GstRTSPMedia * media);
static GstElement *custom_create_pipeline (GstRTSPMediaFactory * factory,
    GstRTSPMedia * media);

static void
start_feed (GstElement * apsrc, guint size, GstRTSPMediaFactory * factory);

static void
stop_feed (GstElement * apsrc, GstRTSPMediaFactory * factory);


G_DEFINE_TYPE (GstRTSPMediaFactoryCustom, gst_rtsp_media_factory_custom, GST_TYPE_RTSP_MEDIA_FACTORY);

static void
gst_rtsp_media_factory_custom_class_init (GstRTSPMediaFactoryCustomClass * klass)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (klass);

  parent_finalize = gobject_class->finalize;

  gobject_class->get_property = gst_rtsp_media_factory_custom_get_property;
  gobject_class->set_property = gst_rtsp_media_factory_custom_set_property;
  gobject_class->finalize = gst_rtsp_media_factory_custom_finalize;

  GST_RTSP_MEDIA_FACTORY_CLASS(klass)->gen_key = custom_gen_key;
  GST_RTSP_MEDIA_FACTORY_CLASS(klass)->get_element = custom_get_element;
  GST_RTSP_MEDIA_FACTORY_CLASS(klass)->construct = custom_construct;
  GST_RTSP_MEDIA_FACTORY_CLASS(klass)->configure = custom_configure;
  GST_RTSP_MEDIA_FACTORY_CLASS(klass)->create_pipeline = custom_create_pipeline;

}

static void
gst_rtsp_media_factory_custom_init (GstRTSPMediaFactoryCustom * factory)
{
  factory->mutex_asrc = g_mutex_new ();
}

static void
gst_rtsp_media_factory_custom_finalize (GObject * obj)
{
  GstRTSPMediaFactoryCustom *factory = GST_RTSP_MEDIA_FACTORY_CUSTOM (obj);

  g_mutex_free (factory->mutex_asrc);

  parent_finalize (obj);
 // G_OBJECT_CLASS (gst_rtsp_media_factory_custom_parent_class)->finalize (obj);
}

static void
gst_rtsp_media_factory_custom_get_property (GObject * object, guint propid,
    GValue * value, GParamSpec * pspec)
{
  GstRTSPMediaFactoryCustom *factory = GST_RTSP_MEDIA_FACTORY_CUSTOM (object);

  switch (propid) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, propid, pspec);
  }
}

static void
gst_rtsp_media_factory_custom_set_property (GObject * object, guint propid,
    const GValue * value, GParamSpec * pspec)
{
  GstRTSPMediaFactoryCustom *factory = GST_RTSP_MEDIA_FACTORY_CUSTOM (object);

  switch (propid) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, propid, pspec);
  }
}

/**
 * gst_rtsp_media_factory_custom_new:
 *
 * Create a new #GstRTSPMediaFactoryCustom instance.
 *
 * Returns: a new #GstRTSPMediaFactoryCustom object.
 */
GstRTSPMediaFactoryCustom *
gst_rtsp_media_factory_custom_new (void)
{
  GstRTSPMediaFactoryCustom *result;

  result = GST_RTSP_MEDIA_FACTORY_CUSTOM (g_object_new (GST_TYPE_RTSP_MEDIA_FACTORY_CUSTOM, NULL));

  return result;
}

static gboolean
compare_media (gpointer key, GstRTSPMedia * media1, GstRTSPMedia * media2)
{
  return (media1 == media2);
}

static void
media_unprepared (GstRTSPMedia * media, GstRTSPMediaFactory * factory)
{
  g_mutex_lock (factory->medias_lock);
  g_hash_table_foreach_remove (factory->medias, (GHRFunc) compare_media, media);
  g_mutex_unlock (factory->medias_lock);

  GstRTSPMediaFactoryCustom* fac = GST_RTSP_MEDIA_FACTORY_CUSTOM(factory);

 // g_mutex_lock (fac->mutex_asrc);
  fac->state = FALSE;

  g_print("media_unprepared\n");
 // g_mutex_lock (fac->mutex_asrc);

}

/**
 * gst_rtsp_media_factory_construct:
 * @factory: a #GstRTSPMediaFactory
 * @url: the url used
 *
 * Prepare the media object and create its streams. Implementations
 * should create the needed gstreamer elements and add them to the result
 * object. No state changes should be performed on them yet.
 *
 * One or more GstRTSPMediaStream objects should be added to the result with
 * the srcpad member set to a source pad that produces buffer of type 
 * application/x-rtp.
 *
 * Returns: a new #GstRTSPMedia if the media could be prepared.
 */
GstRTSPMedia *
gst_rtsp_media_factory_construct (GstRTSPMediaFactory * factory,
    const GstRTSPUrl * url)
{
  gchar *key;
  GstRTSPMedia *media;
  GstRTSPMediaFactoryClass *klass;

  klass = GST_RTSP_MEDIA_FACTORY_GET_CLASS (factory);

  /* convert the url to a key for the hashtable. NULL return or a NULL function
   * will not cache anything for this factory. */
  if (klass->gen_key)
    key = klass->gen_key (factory, url);
  else
    key = NULL;

  g_mutex_lock (factory->medias_lock);
  if (key) {
    /* we have a key, see if we find a cached media */
    media = GST_RTSP_MEDIA (g_hash_table_lookup (factory->medias, key));
    if (media)
      g_object_ref (media);
  } else
    media = NULL;

  if (media == NULL) {
    /* nothing cached found, try to create one */
    if (klass->construct)
      media = klass->construct (factory, url);
    else
      media = NULL;

    if (media) {
      /* configure the media */
      if (klass->configure)
        klass->configure (factory, media);

      /* check if we can cache this media */
      if (gst_rtsp_media_is_shared (media)) {
        /* insert in the hashtable, takes ownership of the key */
        g_object_ref (media);
        g_hash_table_insert (factory->medias, key, media);
        key = NULL;
      }
      if (!gst_rtsp_media_is_reusable (media)) {
        /* when not reusable, connect to the unprepare signal to remove the item
         * from our cache when it gets unprepared */
        g_signal_connect (media, "unprepared", (GCallback) media_unprepared,
            factory);
      }
    }
  }
  g_mutex_unlock (factory->medias_lock);

  if (key)
    g_free (key);

  GST_INFO ("constructed media %p for url %s", media, url->abspath);

  return media;
}

static gchar *
custom_gen_key (GstRTSPMediaFactory * factory, const GstRTSPUrl * url)
{
  gchar *result;
  const gchar *pre_query;
  const gchar *query;

  pre_query = url->query ? "?" : "";
  query = url->query ? url->query : "";

  result =
      g_strdup_printf ("%u%s%s%s", url->port, url->abspath, pre_query, query);

  return result;
}

static GstElement *
custom_get_element (GstRTSPMediaFactory * factory, const GstRTSPUrl * url)
{
  GstElement *element;
  GError *error = NULL;

  GST_RTSP_MEDIA_FACTORY_LOCK (factory);
  /* we need a parse syntax */
  if (factory->launch == NULL)
    goto no_launch;

  /* parse the user provided launch line */
  element = gst_parse_launch (factory->launch, &error);
  if (element == NULL)
    goto parse_error;

  GstRTSPMediaFactoryCustom* fac = GST_RTSP_MEDIA_FACTORY_CUSTOM(factory);


  fac->asrc = gst_bin_get_by_name (GST_BIN (element), "aps");
  fac->state = TRUE;




  GST_RTSP_MEDIA_FACTORY_UNLOCK (factory);

  if (error != NULL) {
    /* a recoverable error was encountered */
    GST_WARNING ("recoverable parsing error: %s", error->message);
    g_error_free (error);
  }
  return element;

  /* ERRORS */
no_launch:
  {
    GST_RTSP_MEDIA_FACTORY_UNLOCK (factory);
    g_critical ("no launch line specified");
    return NULL;
  }
parse_error:
  {
    GST_RTSP_MEDIA_FACTORY_UNLOCK (factory);
    g_critical ("could not parse launch syntax (%s): %s", factory->launch,
        (error ? error->message : "unknown reason"));
    if (error)
      g_error_free (error);
    return NULL;
  }
}

/* try to find all the payloader elements, they should be named 'pay%d'. for
 * each of the payloaders we will create a stream and collect the source pad. */
void
gst_rtsp_media_factory_collect_streams (GstRTSPMediaFactory * factory,
    const GstRTSPUrl * url, GstRTSPMedia * media)
{
  GstElement *element, *elem;
  GstPad *pad;
  gint i;
  GstRTSPMediaStream *stream;
  gboolean have_elem;

  element = media->element;

  have_elem = TRUE;
  for (i = 0; have_elem; i++) {
    gchar *name;

    have_elem = FALSE;

    name = g_strdup_printf ("pay%d", i);
    if ((elem = gst_bin_get_by_name (GST_BIN (element), name))) {
      /* create the stream */
      stream = g_new0 (GstRTSPMediaStream, 1);
      stream->payloader = elem;

      GST_INFO ("found stream %d with payloader %p", i, elem);

      pad = gst_element_get_static_pad (elem, "src");

      /* ghost the pad of the payloader to the element */
      stream->srcpad = gst_ghost_pad_new (name, pad);
      gst_pad_set_active (stream->srcpad, TRUE);
      gst_element_add_pad (media->element, stream->srcpad);
      gst_object_unref (elem);

      /* add stream now */
      g_array_append_val (media->streams, stream);
      have_elem = TRUE;
    }
    g_free (name);

    name = g_strdup_printf ("dynpay%d", i);
    if ((elem = gst_bin_get_by_name (GST_BIN (element), name))) {
      /* a stream that will dynamically create pads to provide RTP packets */

      GST_INFO ("found dynamic element %d, %p", i, elem);

      media->dynamic = g_list_prepend (media->dynamic, elem);

      have_elem = TRUE;
    }
    g_free (name);
  }
}

static GstRTSPMedia *
custom_construct (GstRTSPMediaFactory * factory, const GstRTSPUrl * url)
{
  GstRTSPMedia *media;
  GstElement *element;
  GstRTSPMediaFactoryClass *klass;

  klass = GST_RTSP_MEDIA_FACTORY_GET_CLASS (factory);

  if (!klass->create_pipeline)
    goto no_create;

  if (klass->get_element)
    element = klass->get_element (factory, url);
  else
    element = NULL;
  if (element == NULL)
    goto no_element;

  /* create a new empty media */
  media = gst_rtsp_media_new ();
  media->element = element;

  media->pipeline = klass->create_pipeline (factory, media);
  if (media->pipeline == NULL)
    goto no_pipeline;

  gst_rtsp_media_factory_collect_streams (factory, url, media);

  return media;

  /* ERRORS */
no_create:
  {
    g_critical ("no create_pipeline function");
    return NULL;
  }
no_element:
  {
    g_critical ("could not create element");
    return NULL;
  }
no_pipeline:
  {
    g_critical ("can't create pipeline");
    g_object_unref (media);
    return NULL;
  }
}

static GstElement *
custom_create_pipeline (GstRTSPMediaFactory * factory, GstRTSPMedia * media)
{
  GstElement *pipeline;

  if (media->element == NULL)
    goto no_element;

  pipeline = gst_pipeline_new ("media-pipeline");
  gst_bin_add (GST_BIN_CAST (pipeline), media->element);

  return pipeline;

  /* ERRORS */
no_element:
  {
    g_critical ("no element");
    return NULL;
  }
}

static void
custom_configure (GstRTSPMediaFactory * factory, GstRTSPMedia * media)
{
  gboolean shared, eos_shutdown;
  GstRTSPAuth *auth;

  /* configure the sharedness */
  GST_RTSP_MEDIA_FACTORY_LOCK (factory);
  shared = factory->shared;
  eos_shutdown = factory->eos_shutdown;
  GST_RTSP_MEDIA_FACTORY_UNLOCK (factory);

  gst_rtsp_media_set_shared (media, shared);
  gst_rtsp_media_set_eos_shutdown (media, eos_shutdown);

  if ((auth = gst_rtsp_media_factory_get_auth (factory))) {
    gst_rtsp_media_set_auth (media, auth);
    g_object_unref (auth);
  }
}


GstElement* gst_rtsp_media_factory_custom_get_appsrc(GstRTSPMediaFactoryCustom * factory)
{
 // g_mutex_lock (factory->mutex_asrc);
  return factory->asrc;
 // g_mutex_unlock (factory->mutex_asrc);
}



