#include <pthread.h>
#include <signal.h>
#include <gst/gst.h>
#include <gst/app/gstappsink.h>
#include <gst/rtsp-server/rtsp-server.h>

#include "rtsp-media-factory-custom.h"

static GMainLoop *loop = NULL;
static GstRTSPMediaFactoryCustom *factory1 = NULL;
static GstRTSPMediaFactoryCustom *factory2 = NULL;

static void new_buffer_added (GstAppSink *appsink, gpointer user_data)
{
  GstFlowReturn ret;
  GstElement* appsr = NULL;

  /* pull buffer from appsink of primary pipeline */
  GstBuffer* buffer = gst_app_sink_pull_buffer (appsink);

  //printf("new_buffer_added callback\n ");

  /* push the buffer to auxilary pipes */
  if(factory1)
  {
    if(factory1->state == TRUE)
    {
      //appsr = gst_rtsp_media_factory_custom_get_appsrc(factory1);
      g_debug("pipe1 : BUF PUSHED timestamp[%lli]",GST_BUFFER_TIMESTAMP(buffer));
      g_signal_emit_by_name (factory1->asrc, "push-buffer", buffer, &ret);

    }

  }
  if(factory2)
  {
    if(factory2->state == TRUE)
    {
      //appsr = gst_rtsp_media_factory_custom_get_appsrc(factory2);
      g_debug("pipe2 : BUF PUSHED timestamp[%lli]",GST_BUFFER_TIMESTAMP(buffer));
      g_signal_emit_by_name (factory2->asrc, "push-buffer", buffer, &ret);
    }
  }

  gst_buffer_unref(buffer);
}

void *gst_main (void *arg)
{
  int argc = 0;
  const char *argv[] = {NULL};
  GstElement *pipeline, *sink;
  gchar* primary_pipe = "videotestsrc is-live=true ! video/x-raw-yuv,width=320,height=240 ! appsink name=apsink max-buffers=10 drop=true";

  gst_init (&argc, (char ***) &argv);

  /* Create primary pipeline */
  pipeline = gst_parse_launch(primary_pipe,NULL);
  if (!pipeline)
  {
     g_printerr ("One element could not be created. Exiting.\n");
     return -1;
  }

  sink =  gst_bin_get_by_name (GST_BIN(pipeline), "apsink");

  gst_app_sink_set_max_buffers(GST_APP_SINK(sink),50);
  gst_app_sink_set_drop(GST_APP_SINK(sink),TRUE);

  g_signal_connect (sink, "new-buffer", G_CALLBACK (new_buffer_added), NULL);
  g_object_set (G_OBJECT (sink), "emit-signals", TRUE, NULL);


  /* Set the pipeline to "playing" state*/
  gst_element_set_state (pipeline, GST_STATE_PLAYING);


  /* RTSP server creation and start */
  GstRTSPServer *server;
  server = gst_rtsp_server_new ();

  factory1 = gst_rtsp_media_factory_custom_new ();
  gst_rtsp_media_factory_set_shared(GST_RTSP_MEDIA_FACTORY_CAST(factory1), TRUE);
  gst_rtsp_media_factory_set_launch (GST_RTSP_MEDIA_FACTORY_CAST(factory1), "appsrc name=aps is-live=true block=true !  video/x-raw-yuv,width=320,height=240 ! ffmpegcolorspace ! videoscale ! video/x-raw-yuv,width=320,height=240 ! x264enc ! rtph264pay name=pay0 pt=96");


  factory2 = gst_rtsp_media_factory_custom_new ();
  gst_rtsp_media_factory_set_shared(GST_RTSP_MEDIA_FACTORY_CAST(factory2), TRUE);
  gst_rtsp_media_factory_set_launch (GST_RTSP_MEDIA_FACTORY_CAST(factory2), "appsrc name=aps ! ffmpegcolorspace ! videoscale ! video/x-raw-yuv,width=176,height=120 ! videorate ! video/x-raw-yuv,framerate=10/1 ! ffenc_mpeg4 ! rtpmp4vpay name=pay0 pt=96");

  /* attach the test signals to the URLs */
  GstRTSPMediaMapping *mapping;

  mapping = gst_rtsp_server_get_media_mapping (server);

  gst_rtsp_media_mapping_add_factory (mapping, "/1"  , GST_RTSP_MEDIA_FACTORY_CAST(factory1));
  gst_rtsp_media_mapping_add_factory (mapping, "/2"  , GST_RTSP_MEDIA_FACTORY_CAST(factory2));
  g_object_unref (mapping);

  /* make a mainloop for the default context */
  loop = g_main_loop_new (NULL, FALSE);

  /* attach the server to the default maincontext */
  gst_rtsp_server_attach (server, NULL);

  /* start serving */
  g_main_loop_run (loop);

  return 0;
}

int main (int argc, char *argv[])
{
  // Block all signals for background threads.
  sigset_t new_mask;
  sigfillset(&new_mask);
  sigset_t old_mask;
  pthread_sigmask(SIG_BLOCK, &new_mask, &old_mask);

  // Thread
  pthread_t gst_thread;
  pthread_create(&gst_thread, NULL, gst_main, (void*) NULL);

  // Restore previous signals.
  pthread_sigmask(SIG_SETMASK, &old_mask, 0);

  // Wait for signal indicating time to shut down.
  sigset_t wait_mask;
  sigemptyset(&wait_mask);
  sigaddset(&wait_mask, SIGINT);
  sigaddset(&wait_mask, SIGQUIT);
  sigaddset(&wait_mask, SIGTERM);
  pthread_sigmask(SIG_BLOCK, &wait_mask, 0);

  int sig = 0;
  sigwait(&wait_mask, &sig);

  // stop the thread and wait for it to exit.
  g_main_loop_quit(loop);
  pthread_join(gst_thread, NULL);
}

